;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------
(module unification
	(include "dctp.sch")
	(export (unbind-vars) (unbind-vars-with-count) (ref term) (unify-lists k l) (unify-terms s t)
		(unbind-collect-vars variables) (unbind-last) (unbind-vars-and-undo-var-counter) (occurs-list var termlist)
		(occurs-term var term) (top-level-unify-stop s t))
	(import (display "display.scm") (globals "globals.scm")))

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
  (if (index-variable term) 
      term
      (if (or (non-variable term) (free-variable term) (domain-element term))
	  term
	  (ref (car term)))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind s t))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))


(define (top-level-unify-stop s t)
  (let ((ref-s (ref s)) (ref-t (ref t)))
    (if (and (non-variable ref-s) (non-variable ref-t)
	     (eq? (car ref-s) (car ref-t)))
	(begin
	  (trail-stop)
	  #t)
	(if (and (non-variable ref-s) (variable ref-t))
	    (begin
	      (trail-stop)
	      (bind ref-t ref-s)
	      #t)
	    (if (and (variable ref-s) (non-variable ref-t))
		(begin
		  (trail-stop)
		  (bind ref-s ref-t)
		  #t)
		#f)))))
  
;;;END *********************************************************************
;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin 
       ;;(display "unbinding ") 
       ;;(if (variable (caar trail)) (display "variable "))
       ;;(if (caaar trail)
	   ;;(display-ref (ref (caar trail)))
	   ;;(display (caaar trail)))
       (set-car! (caar trail) '()) 
       ;;(display " to ") 
       ;;(display-ref (caar trail))
       ;;(newline)
       (trail-previous) (unbind-vars))
     (begin (trail-previous) #f)))

(define (unbind-vars-and-undo-var-counter)
;; makes free all variables back to the previous trailstop, decrementing var-counter for each freed variable
 (if (trail-content)
     (begin 
       (set-car! (caar trail) '()) 
       (set! var-counter (- var-counter 1)) 
       (trail-previous) 
       (unbind-vars-and-undo-var-counter))
     (begin (trail-previous) #f)))

(define (unbind-vars-with-count)
;; makes free all variables back to the previous trailstop, returning the number of freed vars
 (if (trail-content)
     (let ((binding (ref (caar trail))))
       (set-car! (caar trail) '()) (trail-previous) 
       (if (variable binding)
	   (unbind-vars-with-count)
	   (+ 1 (unbind-vars-with-count))))
     (begin (trail-previous) 0)))

(define (unbind-last)
;; unbinds the last bound variable on the trail
 (if (trail-content)
     (begin (set-car! (caar trail) '()) (trail-previous))
     (begin (trail-previous) #f)))

(define (unbind-collect-vars variables)
;; makes free and collects all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
	    (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (unbind-collect-vars variables))
     (begin (trail-previous) variables)))

(define (last-binding-variables)
 (let ((old-trail trail) variables)
      (set! variables (last-binding-vars '()))
      (set! trail old-trail)
      variables))

(define (last-binding-vars variables)
 (if (trail-content)
     (begin (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (last-binding-vars variables))
     variables))
